#include <cpu.h>
#include <ints.h>
#include <vm86.h>

extern byte __VM86CodeStart;
extern byte __VM86CodeEnd;
extern byte __Int86;

static const word codeSegment = 0x2000;
static const word stackSegment = 0x8000;

MachineState vm86CPUState;
extern TSS _mainTSS;

static bool vm86EnterHandlerCallback(MachineState *state, void *context)
{
    uintptr_t args[] = { state->EAX, state->EBX, state->ECX, state->EDX, state->ESI, state->EDI, state->EBP };
    //_outsb("vm86EnterHandler\n", 0xE9, 17);
    cpuDisableInterrupts();
    //memcpy(&vm86CPUState, state, sizeof(MachineState));
    size_t size = sizeof(MachineState);
    byte *dst = (byte *)&vm86CPUState;
    byte *src = (byte *)state;
    while(size--) *dst++ = *src++;
    _mainTSS.ESP0 = cpuGetESP();
    _mainTSS.EIP = cpuGetEIP();
    cpuEnterV86(args[1], args[2], args[3], args[4], args[5]);
    return true;
}

static InterruptHandler vm86EnterHandler =
{
    .Callback = vm86EnterHandlerCallback
};

static byte peekb(word seg, word offs)
{
    return *(byte *)((seg << 4) + offs);
}

static word peekw(word seg, word offs)
{
    return *(word *)((seg << 4) + offs);
}

static void pokeb(word seg, word offs, byte val)
{
    *(byte *)((seg << 4) + offs) = val;
}

static void pokew(word seg, word offs, word val)
{
    *(word *)((seg << 4) + offs) = val;
}

static void pushw(MachineState *state, word val)
{
    state->UserSP -= 2;
    pokew(state->UserSS, state->UserSP, val);
}

static word popw(MachineState *state)
{
    word res = peekw(state->UserSS, state->UserSP);
    state->UserSP += 2;
    return res;
}

static bool vm86ExceptionHandlerCallback(MachineState *state, void *context)
{
    // return if not vm86 interrupt
    if(!(state->EFLAGS & (1 << 17)))
        return false;

    byte opcode = peekb(state->CS, state->IP);
    if(opcode == 0xEE)
    {
        outb(state->DX, state->AL);
        ++state->IP;
        return true;
    }
    else if(opcode == 0xEF)
    {
        byte prefix = peekb(state->CS, state->IP - 1);
        if(prefix == 0x66)
            outl(state->DX, state->EAX);
        else
            outw(state->DX, state->AX);
        ++state->IP;
        return true;
    }
    else if(opcode == 0xEC)
    {
        state->AL = inb(state->DX);
        ++state->IP;
        return true;
    }
    else if(opcode == 0xED)
    {
        byte prefix = peekb(state->CS, state->IP - 1);
        if(prefix == 0x66)
            state->EAX = inl(state->DX);
        else
            state->AX = inw(state->DX);
        ++state->IP;
        return true;
    }
    else if(opcode == 0xE6)
    {
        byte port = peekb(state->CS, ++state->IP);
        outb(port, state->AL);
        ++state->IP;
        return true;
    }
    else if(opcode == 0xE7)
    {
        byte prefix = peekb(state->CS, state->IP - 1);
        byte port = peekb(state->CS, ++state->IP);
        if(prefix == 0x66)
            outl(port, state->EAX);
        else
            outw(port, state->AX);
        ++state->IP;
        return true;
    }
    else if(opcode == 0xE4)
    {
        byte port = peekb(state->CS, ++state->IP);
        state->AL = inb(port);
        ++state->IP;
        return true;
    }
    else if(opcode == 0xE5)
    {
        byte prefix = peekb(state->CS, state->IP - 1);
        byte port = peekb(state->CS, ++state->IP);
        if(prefix == 0x66)
            state->EAX = inl(port);
        else
            state->AX = inw(port);
        ++state->IP;
        return true;
    }
    else if(opcode == 0xCD)
    { // int n
        byte intNo = peekb(state->CS, ++state->IP);
        if(intNo == 0xFE)
        { // special interrupt
            //memcpy(state, &vm86CPUState, sizeof(MachineState));
            size_t size = sizeof(MachineState);
            byte *dst = (byte *)state;
            byte *src = (byte *)&vm86CPUState;
            while(size--) *dst++ = *src++;
            return true;
        }
        pushw(state, state->FLAGS);
        state->EFLAGS &= ~((1 << 8) | (1 << 9) | (1 << 18));
        pushw(state, state->CS);
        pushw(state, ++state->IP);
        state->CS = peekw(0x0000, 4 * intNo + 2);
        state->IP = peekw(0x0000, 4 * intNo);
        return true;
    }
    else if(opcode == 0xCF)
    { // iret
        state->IP = popw(state);
        state->CS = popw(state);
        state->FLAGS = popw(state);
        return true;
    }
    else if(opcode == 0x9C)
    { // pushf
        pushw(state, state->FLAGS);
        ++state->IP;
        return true;
    }
    else if(opcode == 0x9D)
    { // popf
        state->FLAGS = popw(state);
        ++state->IP;
        return true;
    }
    else if(opcode == 0xFA)
    { // cli
        state->FLAGS &= ~(1 << 9);
        ++state->IP;
        return true;
    }
    else if(opcode == 0xFB)
    { // sti
        state->FLAGS |= 1 << 9;
        ++state->IP;
        return true;
    }

    _outsb("[vm86] Unknown instruction\n", 0xE9, 27);
    return false;
}

static InterruptHandler vm86ExceptionHandler =
{
    .Callback = vm86ExceptionHandlerCallback
};

void vm86Initialize()
{
    intsRegisterHandler(13, &vm86ExceptionHandler);
    intsRegisterHandler(0xFD, &vm86EnterHandler);

    // install int86 trampoline code in conventional memory
    size_t codeSize = &__VM86CodeEnd - &__VM86CodeStart;
    byte *dst = (byte *)(codeSegment << 4);
    byte *src = &__VM86CodeStart;
    while(codeSize--) *dst++ = *src++;
}

void vm86Enter(word ss, word sp, word cs, word ip, uint32_t arg)
{
    cpuEnterV86Int(0, ss, sp, cs, ip, arg);
}

void vm86Int(byte intNo, VM86Regs *regs)
{
    //memcpy((void *)((codeSegment << 4) + 0x8000), regs, sizeof(VM86Regs));
    size_t size = sizeof(VM86Regs);
    byte *dst = (byte *)((codeSegment << 4) + 0x8000);
    byte *src = (byte *)regs;
    while(size--) *dst++ = *src++;
    vm86Enter(stackSegment, 0x0000, codeSegment, &__Int86 - &__VM86CodeStart, intNo);
}
